'''
Created on Nov 7, 2017

@author: Salim
'''
from lessrpc_common.serialize import Serializer
from lessrpc_common.info.basic import SerializationFormat
from pylodsjson.pylodsjson import JsonObjectMapper, JsonParser
from pylods.deserialize import DeserializationContext




class JsonSerializer(Serializer):
    '''
        JSONSerializer is the default serializer for less RPC if a serializer couldn't be recognized
    '''
    
    __slots__ = ['__mapper']
    
    def __init__(self, mapper=JsonObjectMapper()):
        self.__mapper = mapper
        
    
    def serialize(self, obj, cls, outstream): 
        '''
            serialize into outstream
        :param obj:
        :param cls:
        :param outstream:
        :return no output
        '''
        self.__mapper.write(obj, outstream) 
        
        
    def deserialize(self, instream, cls, ctxt=DeserializationContext.create_context()): 
        '''
            deserialize instream
        :param instream: inputstream
        :param cls class of object to be read:
        :return object instance of class cls
        '''
        parser = JsonParser()
        parser = parser.parse(instream)
        return self.__mapper.read_obj(parser, cls, ctxt=ctxt)
        
    def get_type(self):
        return SerializationFormat("JSON", "")
        

    def prepare(self, module):
        self.__mapper.register_module(module)
    
    
    def copy(self):
        tmp = JsonSerializer()
        tmp.__mapper = self.__mapper.copy()
        return tmp
    
    



'''
Created on Nov 7, 2017

@author: Salim
'''

import traceback
from exceptions import Exception

from io import BytesIO

import cherrypy


from lessrpc_common.types import StatusType
from lessrpc_common.info.basic import SerializationFormat, \
     ServiceInfo, ServiceLocator
from lessrpc_common.errors.less import AcceptTypeHTTPFormatNotParsable, \
    AcceptTypeNotSupported, WrongHTTPMethodException, ContentTypeNotSupported, \
    ContentTypeHTTPFormatNotParsable, ServiceNotSupportedException
from lessrpc_common.info.response import TextResponse, IntegerResponse, \
    ProviderInfoResponse, ServiceSupportResponse, ExecuteRequestResponse, ServiceResponse     
from lessrpc_common.info.request import ServiceRequest

from lessrpc_stub.stubs.base import Stub, BodyWrapper
from lessrpc_common.errors.lessrpc import ServerStubNotInitialized
from pylods.deserialize import DeserializationContext
from lessrpc_stub.stubs.client import NSClient




            
            

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% server stub handler for cherry py
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

class ServerStubHandler():
    '''
        Handles different requests
    '''
    
    def __init__(self, stub, provider):
        self.__stub__ = stub
        self.__provider__ = provider
    
    
    @cherrypy.expose    
    def index(self):
        return "Less RPC service"
    
    @cherrypy.expose
    @cherrypy.config(**{'response.stream': True})
    def ping(self):
        '''
        handles lessrpc's /ping request
        '''
        (flag, _, respserializer, status) = self.process_lessrpc_headers(False)
        
        if not flag:
            return status
        
        try:
            out = BytesIO()
#             out = OutBase64Wrapper(out)
            respserializer.serialize(IntegerResponse(StatusType.OK.value, (1 if self.provider.ping() else 0)), IntegerResponse, out)
            out.flush()
            out.seek(0)
            return out
        except:
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())    
        
        
    
    @cherrypy.expose
    def info(self):
        '''
        handles lessrpc's /info request
        '''
        (flag, _, respserializer, status) = self.process_lessrpc_headers(False)
        
        if not flag:
            return status
              
        try:              
            out = BytesIO()
#             b64 = OutBase64Wrapper(out)
            respserializer.serialize(ProviderInfoResponse(StatusType.OK.value, self.provider.info()), ProviderInfoResponse, out)
            out.flush()
            out.seek(0)        
            return out
        except:
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())
    
    @cherrypy.expose        
    def service(self):
        '''
        handles lessrpc's /support request
        '''
        (flag, reqserializer, respserializer, status) = self.process_lessrpc_headers(True)
        
        if not flag:
            return status
        
        try:
#             out = InBase64Wrapper(BodyWrapper(cherrypy.request.body))
            out = BodyWrapper(cherrypy.request.body)
            service = reqserializer.deserialize(out, ServiceInfo)
        except:
            traceback.print_exc()
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())

        try:
            sup = self.provider.service(service)
        except:
            traceback.print_exc()
            return self.prepare_status(StatusType.SERVICE_NOT_SUPPORTED, respserializer.get_type())
#           
        try:    
            out = BytesIO()
#             b64 = OutBase64Wrapper(out)
            respserializer.serialize(ServiceSupportResponse(StatusType.OK.value, sup), ServiceSupportResponse, out)
            out.flush()
            out.seek(0)        
            return out
        except:
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())
        
    @cherrypy.expose   
    def execute(self):
        '''
        handles lessrpc's /execute request
        '''
        
        (flag, reqserializer, respserializer, status) = self.process_lessrpc_headers(True)
        
        if not flag:
            return status
        
        try:
#             out=InBase64Wrapper(BodyWrapper(cherrypy.request.body))
            out = BodyWrapper(cherrypy.request.body)
            request = reqserializer.deserialize(out, ServiceRequest, ctxt=DeserializationContext.create_context([("CLSLOCATOR", ServiceLocator.create(self.provider.list_services()))]))
        except:
            traceback.print_exc();
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())
        
        if request is None:
            return self.prepare_status(StatusType.INTERNAL_ERROR, respserializer.get_type(), content="Request was null!!")
            
        try:
            result = self.provider.execute(request)
        except ServiceNotSupportedException:
            return self.prepare_status(StatusType.SERVICE_NOT_SUPPORTED, respserializer.get_type())
        except:
            traceback.print_exc();
            return self.prepare_status(StatusType.INTERNAL_ERROR, respserializer.get_type())
        try:
            out = BytesIO()
#             b64 = OutBase64Wrapper(out)
            respserializer.serialize(ExecuteRequestResponse(StatusType.OK.value, ServiceResponse(request.service, result, request.requestid)), ExecuteRequestResponse, out)
            out.flush()
            out.seek(0)
            return out
        except:
            traceback.print_exc();
            return self.prepare_status(StatusType.SERIALIZATION_ERROR, respserializer.get_type())        
        
        
    
    def process_lessrpc_headers(self, inputrequired):
        try:
            # parse serialization format
            try:
                respserializer = self.find_accepted_format();
            except AcceptTypeHTTPFormatNotParsable:
                status = self.prepare_status(StatusType.ACCEPT_TYPE_CANNOT_BE_PARSED, SerializationFormat.default_format())
                return (False, None, None, status);
            except AcceptTypeNotSupported:
                status = self.prepare_status(StatusType.ACCEPT_TYPE_NOT_SUPPORTED, SerializationFormat.default_format())
                return (False, None, None, status);
            # setting response serialization format regardless if there is an error
            # or successful attempt
            cherrypy.response.headers['Content-Type'] = respserializer.get_type().http_format()
            
            # check if get and post are being used properly
            try:
                self.check_http_method_type(inputrequired)
            except WrongHTTPMethodException:
                status = self.prepare_status(StatusType.WRONG_HTTP_METHOD, SerializationFormat.default_format())
                return (False, None, None, status);
                
            # check if content-type exists and it is parsable and supported
            try:
                reqserializer = self.parse_content_type(inputrequired)
            except ContentTypeHTTPFormatNotParsable:
                status = self.prepare_status(StatusType.CONTENT_TYPE_CANNOT_BE_PARSED, SerializationFormat.default_format())
                return (False, None, None, status);
            except ContentTypeNotSupported:
                status = self.prepare_status(StatusType.CONTENT_TYPE_NOT_SUPPORTED, SerializationFormat.default_format())
                return (False, None, None, status);
            
            return (True, reqserializer, respserializer, None)
        
            
            
            
            

        except Exception as e:
            print("Unexpected error has happened!! " + str(e))
            self.prepare_status(StatusType.INTERNAL_ERROR, SerializationFormat.default_format())
    
    
    def find_accepted_format(self):
        accept = cherrypy.request.headers.get("Accept")
        if accept is None or len(accept) == 0:
            raise AcceptTypeHTTPFormatNotParsable(accept)
        # checking accepted formats
        
        formats = None
        try:
            formats = self.parse_accepted_formats(accept)
        except:
            raise AcceptTypeHTTPFormatNotParsable(accept)
        
        if formats is None or len(formats) == 0:
            raise AcceptTypeHTTPFormatNotParsable(accept)
        
        # getting the first supported serializer
        responeformat = self.stub.find_first_accepted_format(formats);
        if responeformat is None:
            raise AcceptTypeNotSupported(accept)
        
        return self.stub.get_serializer(responeformat)  
        
            
        
        
    def parse_accepted_formats(self, txt):
        parts = txt.split(",")
        formats = []
        
        for part in parts:
            formats.append(SerializationFormat.parse_http_format(part))
        
        return formats
    
    
    def parse_content_type(self, inputrequired):
        
        contenttype = cherrypy.request.headers.get("Content-Type")
        requestformat = None
        requestserializer = None
        
        if contenttype is not None and len(contenttype) > 0:
            try:
                requestformat = SerializationFormat.parse_http_format(contenttype)
            except:
                raise ContentTypeHTTPFormatNotParsable(contenttype)
            requestserializer = self.stub.get_serializer(requestformat)
            if requestserializer is None:
                raise ContentTypeNotSupported(requestformat)
        else:
            if inputrequired:
                raise ContentTypeHTTPFormatNotParsable(contenttype)
            
        
        return requestserializer
    
    
    def prepare_status(self, status, frmt=None, content=None):
        if frmt is not None:
            cherrypy.response.headers['Content-Type'] = frmt.http_format()
            
        if content is None:
            content = status.name

        serializer = self.stub.get_serializer(frmt)

        out = BytesIO()        
#         b64 = OutBase64Wrapper(out)
        serializer.serialize(TextResponse(status.value, content), TextResponse, out)
        out.flush()
        return out.getvalue()
        
        
        
        
    
    def check_http_method_type(self, inputrequired):
        if inputrequired and cherrypy.request.method.upper() != 'POST':
            raise WrongHTTPMethodException(cherrypy.request.method.upper())
        
        if not inputrequired and cherrypy.request.method.upper() == 'POST':
            raise WrongHTTPMethodException(cherrypy.request.method.upper())
            
    

    
    def __get_stub__(self):
        return self.__stub__;
    
    stub = property(__get_stub__)
    
    def __get_provider__(self):
        return self.__provider__;
    
    provider = property(__get_provider__)






# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% server stub 
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class ServerStub(Stub):
    '''
    basic ServerStub class for lessRPC 
    '''
    
    
    def __init__(self, provider, serializers=[]):
        Stub.__init__(self, serializers)
        self.__port__ = provider.info().port
        self.__provider__ = provider
        
        
    
    
    def start(self):
        if self.__provider__ is None:
            raise ServerStubNotInitialized()
        
        self.before_start();
        
        
        
        cherrypy.config.update({'server.socket_port': self.__port__})
        cherrypy.tree.mount(ServerStubHandler(self, self.__provider__))
        cherrypy.engine.start()
        cherrypy.engine.block()
        
        self.after_start()
    
    
    def before_start(self):
        '''
            calls anything needs to be executed before starting
        '''
        pass
    
    def after_start(self):
        '''
            writes anything that needs to be executed after already started
        '''
        pass
    
    
    def stop(self):
        '''
        stop the server
        '''
        self.before_stop();
        cherrypy.engine.exit()
        self.after_stop();
    
    
    def before_stop(self):
        '''
            runs what needs to be executed before stopping the server
        '''
        pass
    
    def after_stop(self):
        '''
            runs anything needed to be executed after server has stopped
        '''
        pass
    
    
    def __get_port(self):
        return self.__port__
    
    def __get_provider(self):
        return self.__provider__
    
    
    
    def find_first_accepted_format(self, formats):
        '''
            finds the first accepted format from a list of formats
        :param formats: list of SerializationFormat
        :return format: an instance of SerializationFormat
        '''
        if formats is None or len(formats) == 0:
            return None
        
        for frmt in formats:
            if(self.accepts(frmt)):
                return frmt
        
        return None
    
    
    port = property(__get_port)
    provider = property(__get_provider)
    
    

class NSServerStub(ServerStub):
    
    def __init__(self, provider, nsinfo, serializers=[]):
        ServerStub.__init__(self, provider, serializers)
        self.__nsinfo__ = nsinfo
        self.__ns__ = NSClient(nsinfo, serializers)
        

    def after_start(self):
        '''
            Registers all supported services for this provider
        '''
        for support in self.provider.list_support():
            self.ns.register(support)
        
        
    def before_stop(self):
        '''
            unregisters all services just before stopping the server stub
        '''
        self.ns.unregister_all(self.provider.info())
        
        
        
        
    
    def __get_nsinfo(self):
        return self.__nsinfo__
    
    def __get_ns(self):
        return self.__ns__

    
    nsinfo = property(__get_nsinfo)
    ns = property(__get_ns)    






    
    
    






'''
Created on Jan 2, 2018

@author: Salim
'''
class NoProviderAvailableException(Exception):
    
    def __init__(self, service):
        # Call the base class constructor with the parameters it needs
        super(NoProviderAvailableException, self).__init__("Tried to fetch a provider info from NameServer but NameServer did not return any for service: "
                + service)


'''
Created on Nov 7, 2017

@author: Salim
'''

import random
from datetime import datetime
from lessrpc_stub.serializer import JsonSerializer
import base64
from _io import BytesIO







class Stub():
    
    __slots__ = ['__serializers', '__serializer_map', '__accepted_type_string']
    
    
    def __init__(self, serializers=[]):
        self.__serializers = serializers
        self.__serializer_map = {}

        hasjson = False
        
        # intialize serializers map
        for serializer in serializers:
            self.__serializer_map[serializer.get_type()] = serializer
            if serializer.get_type().name.lower == "json":
                hasjson = True
            
        if not hasjson:
            json = JsonSerializer()
            self.__serializers.append(json)
            self.__serializer_map[json.get_type()] = json
    

    def get_serializer(self, fmt):
        return self.__serializer_map.get(fmt, None)
    
    
    def get_serializers(self):
        return self.__serializers
    
    
    def get_accepted_types(self, accept=None):
        
        if accept is None:
            accept = []
            for s in self.get_serializers():
                accept.append(s.get_type())
                
                
        txt = ""
        for frmt in accept:
            if len(txt) > 0:
                txt = txt + " , "
            txt = txt + frmt.http_format()
        
        return txt;
    

    def get_random_id(self):
        random.seed(datetime.now())
        return random.randint(0, 2 ** 31 - 1)
    
    
    def accepts(self, frmt):
        return frmt in self.__serializer_map
    
    serializers = property(get_serializers)
    
    
class BodyWrapper():
    
    def __init__(self, body): 
        self.body = body
        self.cache = None
        
        
    def read(self, size=None):
        
            
        if size is not None and size == 0:
            if self.cache is None:
                self.cache = self.body.read(1)
            else:
                self.cache = self.cache + self.body.read(1)  
            return self.cache[0:0]
        else:
            data = None
            if size is None:
                data = self.body.read()
            else:
                data = self.body.read(size)
            
            if self.cache is not None:
                data = self.cache + data;
                self.cache = None
            
            return data    
    
 
class InBase64Wrapper():
    
    DO_BASE64 = False
    
    def __init__(self, instream):
        self.instream = instream
        self.cache = ''
        
        
     
    def read(self, size=-1):
        if self.DO_BASE64 :
        
            data = self.instream.read(size)
            finished = len(data) < size or size == -1
            
            if finished:
                return base64.b64decode(self.cache + data)
            else:
                idx = data.rfind('\n')
                idx = max(idx, data.rfind('='))
                if idx > -1:
                    out = base64.b64decode(self.cache + data[0:idx])
                    self.cache = data[idx:]
                    return out
                else:
                    self.cache = self.cache + data
        else:
            return self.instream.read(size)
        
    
    def close(self):
        self.instream.close()
            
    
class OutBase64Wrapper():
    
    
    DO_BASE64 = False
    
    def __init__(self, outstream):
        self.outstream = outstream
        self.cache = ''


    
    def write(self, data):
        if self.DO_BASE64 :
            size = len(data) + len(self.cache)
            sizetowrite = int(round(size / 3))*3
            fromcache = min(sizetowrite, len(self.cache))
            out = self.cache[0:fromcache] + data[0:(sizetowrite - fromcache)]
            self.cache = self.cache[fromcache:len(self.cache)] + data[(sizetowrite - fromcache):len(data)]
            self.outstream.write(base64.b64encode(out))
        else:
            self.outstream.write(data)
        

    
    def flush(self):
        if self.DO_BASE64 :
            if len(self.cache) > 0:
                self.outstream.write(base64.b64encode(self.cache))
                self.outstream.flush()
        else:
            self.outstream.flush()
    
    def close(self):
        self.flush()
        self.outstream.close()
    
    

# out = BytesIO()
# out.write(base64.b64encode("test-ts-xx-yyy-zzzz-mmmmm"))
# out.seek(0)
# # 
# b64= InBase64Wrapper(out)
# 
# print(b64.read(2))
# print(b64.read(4))
# print(b64.read(1))
# print(b64.read(10))
# print(b64.read(7))
# print(b64.read(1))
# print(b64.read(1))




# out = BytesIO()
# # 
# b64 = OutBase64Wrapper(out)
# b64.write("1234"); print(out.getvalue())
# b64.write("5"); print(out.getvalue())
# b64.write("67"); print(out.getvalue())
# b64.write(""); print(out.getvalue())
# b64.flush(); print(out.getvalue())


#



from setuptools import setup

setup(
  name = 'lessrpc_stub',
  packages = ['lessrpc_stub','lessrpc_stub.stubs'], # this must be the same as the name above
  version = '1.0.15',
  description = 'Less RPC stubs project',
  author = 'Salim Malakouti',
  author_email = 'salim.malakouti@gmail.com',
  license = 'MIT',
  url = 'https://github.com/LessRPC/lessrpc_stub_py2x/', # use the URL to the github repo
  download_url = 'https://github.com/LessRPC/lessrpc_stub_py2x/archive/1.0.15.tar.gz', # I'll explain this in a second
  keywords = ['python','serialization','deserialization','rpc','rmi','less rpc', 'client', 'server'], # arbitrary keywords
  classifiers = ['Programming Language :: Python :: 2.7'],
  install_requires=['pylods','pylods-json','httpoutputstream','cherrypy','enum34' ],
  python_requires='~=2.7'
)
